// =====================================================================================================================
//  fennec, a free and open source game engine
//  Copyright © 2025  Medusa Slockbower
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <https://www.gnu.org/licenses/>.
// =====================================================================================================================

///
/// \file types.h
/// \brief basic types of the c++ language
///
///
/// \details
/// \author Medusa Slockbower
///
/// \copyright Copyright © 2025 Medusa Slockbower ([GPLv3](https://www.gnu.org/licenses/gpl-3.0.en.html))
///
///

#ifndef FENNEC_LANG_TYPES_H
#define FENNEC_LANG_TYPES_H

#include <cstdint>

namespace fennec
{

// Basic Types =========================================================================================================

///
/// \name Basic Types
/// @{

///
/// \brief A conditional type
using bool_t = bool;

///
/// \brief A type capable of holding an ascii value
using char_t  = char;

///
/// \brief A type with the size of a char, capable of holding a signed 8-bit integer
using schar_t = signed char;

///
/// \brief A type with the size of a char, capable of holding an unsigned 8-bit integer
using uchar_t = unsigned char;

///
/// \brief A signed short type, capable of holding signed 16-bit integer
using short_t = signed short;

///
/// \brief An unsigned short type, capable of holding an unsigned signed 16-bit integer
using ushort_t = unsigned short;

///
/// \brief A signed integer type, size varies by implementation, but typically 32-bit
using int_t = signed int;

///
/// \brief An unsigned integer type, size varies by implementation, but typically 32-bit
using uint_t = unsigned int;

///
/// \brief A signed integer type, with a size of at least 32-bits
using long_t  = signed long;

///
/// \brief An unsigned integer type, with a size of at least 32-bits
using ulong_t = unsigned long;

///
/// \brief A signed integer type, with a size of 64-bits
using llong_t  = signed long long;

///
/// \brief An unsigned integer type, with a size of 64-bits
using ullong_t = unsigned long long;

///
/// \brief A single-precision floating-point type, typically with a size of 32-bits
using float_t = float;

///
/// \brief A double-point type, typically with a size of 64-bits
using double_t = double;
/// @}



// Sized Arithmetic Types ==============================================================================================

///
/// \name Sized Integer Types
/// @{

///
/// \brief Signed  8-bit integer
using int8_t = schar_t;

///
/// \brief Signed 16-bit integer
using int16_t = short_t;

///
/// \brief Signed 32-bit integer
using int32_t =  long_t;

///
/// \brief Signed signed 64-bit integer
using int64_t = llong_t;

/// @}



///
/// \name Sized Unsigned Integer Types
/// @{

///
/// \brief Unsigned 8-bit integer
using uint8_t   = uchar_t;

///
/// \brief Unsigned 16-bit integer
using uint16_t  = ushort_t;

///
/// \brief Unsigned 32-bit integer
using uint32_t  =  ulong_t;

///
/// \brief Unsigned 64-bit integer
using uint64_t  = ullong_t;

/// @}



///
/// \name Sized Floating-Point Types
/// @{

///
/// \brief A single-precision floating-point scalar
using float32_t = float_t;

///
/// \brief A double-precision floating-point scalar
using float64_t = double_t;
/// @}



///
/// \name Special Types
/// @{

///
/// \brief Null Pointer Type
using nullptr_t = decltype(nullptr);

///
/// \brief Signed Integer Capable of Holding a Pointer to void
using intptr_t  = intptr_t;

///
/// \brief Unsigned Integer Capable of Holding a Pointer to void
using uintptr_t = uintptr_t;

///
/// \brief Maximum Width Signed Integer Type
using intmax_t  = __INTMAX_TYPE__;

///
/// \brief Maximum Width Unsigned Integer Type
using uintmax_t = __UINTMAX_TYPE__;

///
/// \brief Unsigned Integer Type Returned By `sizeof`, `sizeof...`, and `alignof`
using size_t    = __SIZE_TYPE__;

///
/// \brief Signed Integer Type Returned by the Subtraction of two Pointers
using ptrdiff_t = __PTRDIFF_TYPE__;

///
/// \brief undefined class for SFINAE
class undefined_t;

///
/// \brief Void type used for SFINAE
template<typename...>
using void_t = void;

/// @}

}

#endif // FENNEC_LANG_TYPES_H
